SystemOrganization addCategory: #Unification!
SystemOrganization addCategory: #'Unification-Tests'!

Dictionary subclass: #MostGeneralUnifier
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Unification'!

!MostGeneralUnifier commentStamp: 'fbs 1/4/2012 10:31' prior: 0!
Given two structures s and t, each containing zero or more MetaVariables, to what values must those variables be set for s and t to be structurally equivalent?

If s := 1 and t := #x asVariable, then the most general unifier (mgu) will be {#x asVariable -> 1}.

If s := 2 and t := 1 then the mgu does not exist.

If s := t := 1 then the mgu is {}, or "there's no unification needed".!

!Symbol methodsFor: '*Unification' stamp: 'fbs 7/22/2011 13:18'!
asVariable
	^ MetaVariable named: self.! !

!SequenceableCollection methodsFor: '*Unification' stamp: 'fbs 1/4/2012 11:33'!
findSolutionWithUnifier: anMGU in: solutionFinder
	| newUnifier |
	newUnifier := anMGU copy.
	self do:
		[:each | | d |
		d := solutionFinder	findSolution: each using: newUnifier.
		newUnifier addAll: d].
	^ newUnifier.! !

!SequenceableCollection methodsFor: '*Unification' stamp: 'fbs 1/3/2012 15:30'!
unificationClosureWith: anObject in: termRelation
	anObject isMetaVariable
		ifTrue: [^ termRelation union: self with: anObject].
		
	^ anObject unificationClosureWithSequence: self in: termRelation.! !

!SequenceableCollection methodsFor: '*Unification' stamp: 'fbs 1/3/2012 16:14'!
unificationClosureWithSequence: aCollection in: termRelation
	| rel |
	(self size = aCollection size) ifFalse: [UnificationFailure signal].
	rel := termRelation.
	self with: aCollection do:
		[:si :ti | rel := (rel find: si) unificationClosureWith: (rel find: ti) in: rel].
	^ rel.! !

Error subclass: #UnificationFailure
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Unification'!

!UnificationFailure commentStamp: 'fbs 5/24/2011 17:35' prior: 0!
I represent the inability to unify two patterns.!

TestCase subclass: #MatchingTest
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Unification-Tests'!

!MatchingTest methodsFor: 'helpers' stamp: 'fbs 1/9/2012 09:38'!
assert: left equals: right description: msg
	self
		assert: left = right
		description: ('{1}, expected {2}, got {3}' format: {msg. left printString. right printString.}).! !

!MatchingTest methodsFor: 'testing - helpers' stamp: 'fbs 1/9/2012 09:38'!
testAssertEqualsDescriptionGivesMeaningfulErrorMessage
	[self assert: 1 equals: 2 description: 'Fail!!'] on: TestFailure do:
		[ :failure |
		self assert: 'Fail!!, expected 1, got 2' equals: failure messageText].! !

!MatchingTest methodsFor: 'testing - matching' stamp: 'fbs 7/28/2011 14:18'!
testEmptyMguPassedIntoValueBlockMatchWins
	| result substitution |
	result := 1 matchOneOf: {[1] -> [:mgu | substitution := mgu. 'foo']} otherwise: ['bar'].
	self assert: result = 'foo'.
	self assert: substitution isEmpty.! !

!MatchingTest methodsFor: 'testing - matching' stamp: 'fbs 1/5/2012 09:34'!
testFirstMatchWins
	| result |
	result := 1
		matchOneOf:
			{[1] -> ['first match'].
			[1] -> ['second match'].}
		otherwise: ['otherwise block'].
	self assert: result = 'first match' description: 'Unexpected match invoked'.! !

!MatchingTest methodsFor: 'testing - matching' stamp: 'fbs 1/5/2012 09:35'!
testMatchPassesUnifierIntoBlock
	| result mgu |
	result := {1}
		matchOneOf:
			{[{#x asVariable}] -> [:vars | mgu := vars. 'match']}
		otherwise: ['otherwise block'].
	self assert: result = 'match' description: 'No match found'.
	self assert: mgu notNil description: 'MGU not passed into block'.
	self assert: (mgu at: #x asVariable) = 1 description: 'Unexpected MGU'.! !

!MatchingTest methodsFor: 'testing - matching' stamp: 'fbs 1/5/2012 09:36'!
testMatchingTriesAllBlocks
	| result |
	result := 1
		matchOneOf:
			{[0] -> ['first match'].
			[1] -> ['second match'].}
		otherwise: ['otherwise block'].
	self assert: result = 'second match' description: 'Unexpected result'.! !

!MatchingTest methodsFor: 'testing - matching' stamp: 'fbs 1/5/2012 09:32'!
testMatchingUnifiesBidirectionally
	| result mgu |
	result := {#x asVariable. 1}
		matchOneOf:
			{[{2. #y asVariable}] -> [:vars | mgu := vars. 'match']}
		otherwise: ['otherwise block'].
	self assert: result = 'match' description: 'No match found'.
	self assert: mgu notNil description: 'MGU not passed into block'.
	self assert: (mgu at: #x asVariable) equals: 2 description: 'Unexpected MGU (x)'.
	self assert: (mgu at: #y asVariable) equals: 1 description: 'Unexpected MGU (y)'.! !

!MatchingTest methodsFor: 'testing - matching' stamp: 'fbs 1/5/2012 09:33'!
testNoMatchEvaluatesOtherwiseBlock
	| result |
	result := 1
		matchOneOf:
			{['not a 1'] -> ['matched something']}
		otherwise: ['otherwise block'].
	self assert: result = 'otherwise block' description: 'Otherwise block not invoked'.! !

!MatchingTest methodsFor: 'testing - matching' stamp: 'fbs 1/5/2012 09:33'!
testNoMatchEvaluatesRaisesError
	self should: [1
		matchOneOf:
			{['not a 1'] -> ['matched something']}]
		raise: Error.! !

TestCase subclass: #MetaVariableTest
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Unification-Tests'!

!MetaVariableTest methodsFor: 'testing - comparing' stamp: 'fbs 1/4/2012 20:34'!
testMetaVariableEqualToSameNamedMetaVariable
	self assert: #x asVariable equals: #x asVariable.! !

!MetaVariableTest methodsFor: 'testing - comparing' stamp: 'fbs 1/4/2012 20:35'!
testMetaVariableHashEqualToSameNamedMetaVariableHash
	self assert: #x asVariable hash equals: #x asVariable hash.! !

!MetaVariableTest methodsFor: 'testing - comparing' stamp: 'fbs 1/4/2012 20:36'!
testMetaVariableHashNotEqualToDifferentlyNamedMetaVariableHash
	self deny: #x asVariable hash = #X asVariable hash.! !

!MetaVariableTest methodsFor: 'testing - comparing' stamp: 'fbs 1/4/2012 20:35'!
testMetaVariableNotEqualToDifferentlyNamedMetaVariable
	self deny: #x asVariable = #X asVariable.! !

!MetaVariableTest methodsFor: 'testing - comparing' stamp: 'fbs 1/4/2012 20:35'!
testMetaVariableNotEqualToNil
	self deny: #x asVariable = nil! !

!MetaVariableTest methodsFor: 'testing' stamp: 'fbs 5/24/2011 11:03'!
testUnapplyShouldDestructureCorrectly
	self assert: (MetaVariable named: #a) unapply = {#Variable. #a.}! !

!MetaVariableTest methodsFor: 'testing' stamp: 'fbs 5/24/2011 22:08'!
testVariableHasNoChildren
	self deny: (MetaVariable named: #foo) hasChildren.

	self assert: (MetaVariable named: #foo) children isEmpty.! !

!MetaVariableTest methodsFor: 'testing' stamp: 'fbs 5/25/2011 22:32'!
testVariableIsVariable
	self assert: (MetaVariable named: #foo) isMetaVariable.! !

TestCase subclass: #OrUnifierTest
	instanceVariableNames: 'matcher'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Unification-Tests'!

!OrUnifierTest methodsFor: 'running' stamp: 'fbs 10/7/2011 07:45'!
setUp
	matcher := OrUnifier
		left: (TreeNode left: #x asVariable)
		right: (TreeNode right: #x asVariable).! !

!OrUnifierTest methodsFor: 'test' stamp: 'fbs 10/7/2011 07:45'!
testFirstOptionUnifies
	| mgu |
	mgu := matcher =? (TreeNode left: (Leaf value: 1)).
	self assert: (mgu at: (#x asVariable)) = (Leaf value: 1).! !

!OrUnifierTest methodsFor: 'test' stamp: 'fbs 8/9/2011 07:30'!
testNoUnificationPossibleStillSignalsError
	self should: [matcher =? Tree empty] raise: UnificationFailure! !

!OrUnifierTest methodsFor: 'test' stamp: 'fbs 10/7/2011 07:45'!
testSecondOptionUnifies
	| mgu |
	mgu := matcher =? (TreeNode right: (Leaf value: 1)).
	self assert: (mgu at: (#x asVariable)) = (Leaf value: 1).! !

!OrUnifierTest methodsFor: 'test' stamp: 'fbs 12/28/2011 09:30'!
testUnifiesAsRightHandOperandAgainstFirstOption
	| mgu |
	mgu := (TreeNode left: (Leaf value: 1)) =? matcher.
	self assert: (mgu at: (#x asVariable)) = (Leaf value: 1).! !

!OrUnifierTest methodsFor: 'test' stamp: 'fbs 12/28/2011 09:30'!
testUnifiesAsRightHandOperandAgainstSecondOption
	| mgu |
	mgu := (TreeNode right: (Leaf value: 1)) =? matcher.
	self assert: (mgu at: (#x asVariable)) = (Leaf value: 1).! !

TestCase subclass: #SolutionFinderTest
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Unification-Tests'!

!SolutionFinderTest commentStamp: 'fbs 1/4/2012 10:39' prior: 0!
Most of SolutionFinder's tests are found in UnificationTest.!

!SolutionFinderTest methodsFor: 'fixtures' stamp: 'fbs 1/5/2012 09:28'!
makePartition
	| partition |
	partition := VariableTrackingUnionFind usingArrayType: PersistentCollection partitioning: Dictionary new.
	partition := (partition find: 1) unificationClosureWith: (partition find: #x asVariable) in: partition.
	^ partition.! !

!SolutionFinderTest methodsFor: 'testing - instance creation' stamp: 'fbs 1/5/2012 09:29'!
testCanCreateFinder
	| partition |
	partition := self makePartition.
	SolutionFinder findSolutionFor: partition starting: 1.! !

TestCase subclass: #UnificationMatcherTest
	instanceVariableNames: 'matcher substitution'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Unification-Tests'!

!UnificationMatcherTest methodsFor: 'helpers' stamp: 'fbs 7/28/2011 22:17'!
assert: left equals: right description: msg
	self
		assert: left = right
		description: ('{1} , expected {2}, got {3}' format: {msg. left printString. right printString.}).! !

!UnificationMatcherTest methodsFor: 'running' stamp: 'fbs 10/7/2011 07:44'!
setUp
	matcher := UnificationMatcher matcherUsing:
		{[Leaf value: 1] -> ['first leaf'].
		[Leaf value: 1] -> ['second leaf'].
		[Tree empty] -> ['tree'].
		[TreeNode left: #x asVariable] -> [:mgu | substitution := mgu. 'mgu-using case']}
		otherwise: ['otherwise block'].! !

!UnificationMatcherTest methodsFor: 'testing - matching' stamp: 'fbs 8/8/2011 19:09'!
testFirstMatchWins
	| result |
	result := matcher match: (Leaf value: 1).
	self assert: result = 'first leaf' description: 'Unexpected match invoked'.! !

!UnificationMatcherTest methodsFor: 'testing - matching' stamp: 'fbs 10/7/2011 07:44'!
testMatchPassesUnifierIntoBlock
	| result |
	result := matcher match: (TreeNode left: #y asVariable).
	
	self assert: result = 'mgu-using case' description: 'No match found'.
	self assert: substitution notNil description: 'MGU not passed into block'.
	self assert: (substitution at: #y asVariable) = (#x asVariable) description: 'Unexpected MGU'.! !

!UnificationMatcherTest methodsFor: 'testing - matching' stamp: 'fbs 8/8/2011 20:26'!
testMatchingTriesAllBlocks
	| result |
	result := matcher match: Tree empty.
	self assert: result = 'tree' description: 'Unexpected result'.! !

!UnificationMatcherTest methodsFor: 'testing - matching' stamp: 'fbs 8/8/2011 20:26'!
testNoMatchEvaluatesOtherwiseBlock
	| result |
	result := matcher match: (Leaf value: 2).
	self assert: result = 'otherwise block' description: 'Otherwise block not invoked'.! !

!UnificationMatcherTest methodsFor: 'testing - matching' stamp: 'fbs 7/28/2011 22:28'!
testNoMatchEvaluatesRaisesError
	| m |
	m := UnificationMatcher matcherUsing:
			{[Tree empty] -> ['matched something']}.
			
	self should: [m matchAgainst: (Leaf value: 1)] raise: Error.! !

TestCase subclass: #UnificationOfAlgebraicDataTypeTest
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Unification-Tests'!

!UnificationOfAlgebraicDataTypeTest methodsFor: 'testing' stamp: 'fbs 1/5/2012 09:29'!
testAlgebraicDataTypeOrReturnsOrMatcher
	| mgu |
	mgu := ((TreeNode left: #x asVariable) or: (TreeNode right: #x asVariable)) =? (TreeNode right: (Leaf value: 1)).
	self assert: (mgu at: (#x asVariable)) = (Leaf value: 1).! !

!UnificationOfAlgebraicDataTypeTest methodsFor: 'testing' stamp: 'fbs 1/5/2012 09:24'!
testDifferentArityTermsShouldNotUnify
	"f(x) ~= g(x,y)"
	|f g |
	f := VariadicTerm with: #(#x).
	g := VariadicTerm with: #(#x #y).
	self should: [f unifyWith: g] raise: UnificationFailure.! !

!UnificationOfAlgebraicDataTypeTest methodsFor: 'testing - collection unification' stamp: 'fbs 1/9/2012 09:34'!
testUnappliedAlgebraicDataTypesCantUnifyWithAlgebraicDataTypes
	self shouldnt: [ (Leaf value: 1) =? #(Leaf 1) ] raise: UnificationFailure.! !

!UnificationOfAlgebraicDataTypeTest methodsFor: 'testing - collection unification' stamp: 'fbs 1/5/2012 09:25'!
testUnificationWithCollectionsWorksBidirectionally
	| mgu |
	mgu := {#x asVariable. 1.} =? {VariadicTerm with: {#a asVariable}. #y asVariable.}.
	self assert: (VariadicTerm with: {#a asVariable}) equals: (mgu at: #x asVariable).
	self assert: 1 equals: (mgu at: #y asVariable).! !

TestCase subclass: #UnificationOfCollectionsTest
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Unification-Tests'!

!UnificationOfCollectionsTest methodsFor: 'testing' stamp: 'fbs 1/5/2012 09:26'!
testNormalObjectsDontUnifyWithCollections
	self should: [1 =? #()] raise: UnificationFailure.
	self should: [#() =? 1] raise: UnificationFailure.! !

!UnificationOfCollectionsTest methodsFor: 'testing' stamp: 'fbs 1/5/2012 09:26'!
testUnificationWorksBidirectionally
	| mgu |
	mgu := {#x asVariable. 1.} =? {2. #y asVariable}.
	self assert: 2 equals: (mgu at: #x asVariable).
	self assert: 1 equals: (mgu at: #y asVariable).! !

!UnificationOfCollectionsTest methodsFor: 'testing' stamp: 'fbs 10/7/2011 07:57'!
testUnificationWorksLeftward
	| mgu |
	mgu := {#x asVariable. #y asVariable.} =? #(1 (2 3)).
	self assert: 1 equals: (mgu at: #x asVariable).
	self assert: #(2 3) equals: (mgu at: #y asVariable).! !

!UnificationOfCollectionsTest methodsFor: 'testing' stamp: 'fbs 10/7/2011 07:56'!
testUnificationWorksRightward
	| mgu |
	mgu := #(1 (2 3)) =? {#x asVariable. #y asVariable.}.
	self assert: 1 equals: (mgu at: #x asVariable).
	self assert: #(2 3) equals: (mgu at: #y asVariable).! !

TestCase subclass: #UnificationTest
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Unification-Tests'!

!UnificationTest methodsFor: 'testing' stamp: 'fbs 1/8/2012 22:49'!
testAnythingCanUnifyAgainstMetaVariables
	self shouldnt: ['a' =? #x asVariable] raise: UnificationFailure.
	self shouldnt: [$a =? #x asVariable] raise: UnificationFailure.
	self shouldnt: [1 =? #x asVariable] raise: UnificationFailure.
	self shouldnt: [#() =? #x asVariable] raise: UnificationFailure.
	self shouldnt: [Object new =? #x asVariable] raise: UnificationFailure.! !

!UnificationTest methodsFor: 'testing' stamp: 'fbs 1/4/2012 22:16'!
testCyclicStructuresCantUnify
	| y |
	y := MetaVariable named: #y.
	self should: [y unifyWith: {y}] raise: UnificationFailure.! !

!UnificationTest methodsFor: 'testing' stamp: 'fbs 1/4/2012 22:16'!
testImpossibleUnificationShouldNotUnify
	self should: [1 unifyWith: 2] raise: UnificationFailure.! !

!UnificationTest methodsFor: 'testing' stamp: 'fbs 1/8/2012 22:56'!
testMetaVariablesCanUnifyAgainstAnything
	self shouldnt: [#x asVariable =? 'a'] raise: UnificationFailure.
	self shouldnt: [#x asVariable =? $a] raise: UnificationFailure.
	self shouldnt: [#x asVariable =? 1] raise: UnificationFailure.
	self shouldnt: [#x asVariable =? #()] raise: UnificationFailure.
	self shouldnt: [#x asVariable =? Object new] raise: UnificationFailure.! !

!UnificationTest methodsFor: 'testing' stamp: 'fbs 5/29/2011 15:08'!
testSelfUnificationShouldUnify
	| x |
	x := MetaVariable named: #x.
	"Tautologies"
	self assert: (1 unifyWith: 1) isEmpty.
	self assert: (x unifyWith: x) isEmpty.! !

!UnificationTest methodsFor: 'testing' stamp: 'fbs 5/26/2011 07:43'!
testTrivialBidirectionalUnificationShouldUnify
	| matches x y |
	x := MetaVariable named: #x.
	y := MetaVariable named: #y.
	matches := x unifyWith: y.
	self assert: matches = (Dictionary newFrom: {x -> y}).! !

!UnificationTest methodsFor: 'testing' stamp: 'fbs 1/4/2012 22:18'!
testTrivialUnificationsShouldUnify
	| x |
	x := MetaVariable named: #x.
	self assert: (x unifyWith: 2) = (Dictionary newFrom: {x -> 2}).
	self assert: (2 unifyWith: x) = (Dictionary newFrom: {x -> 2}).! !

!UnificationTest methodsFor: 'testing' stamp: 'fbs 1/4/2012 22:22'!
testTwoVariableBidirectionalUnificationShouldUnifyBoth
	| matches left right |
	left := MetaVariable named: #left.
	right := MetaVariable named: #right.
	matches := {left. 2} unifyWith: {1. right}.
	self assert: matches = (Dictionary newFrom: {
		left -> 1.
		right -> 2}).! !

!UnificationTest methodsFor: 'testing' stamp: 'fbs 1/4/2012 22:21'!
testTwoVariableUnificationShouldUnifyBoth
	| matches left right |
	left := MetaVariable named: #left.
	right := MetaVariable named: #right.
	matches := {left. right} unifyWith: {1. 2}.
	self assert: matches = (Dictionary newFrom: {
		left -> 1.
		right -> 2}).! !

!UnificationTest methodsFor: 'testing' stamp: 'fbs 1/5/2012 09:27'!
testUnificationWorksOnDeepStructures
	| matches x y |
	x := MetaVariable named: #x.
	y := MetaVariable named: #y.
	matches := { {{x}} } unifyWith: {y}.
	self assert: matches = (Dictionary newFrom: { y -> {{x}} }).! !

!UnificationTest methodsFor: 'testing' stamp: 'fbs 1/4/2012 22:24'!
testVariableUnifyingAgainstTwoVariables
	| matches x y z |
	x := MetaVariable named: #x.
	y := MetaVariable named: #y.
	z := MetaVariable named: #z.
	matches := {x. x.} unifyWith: {y. z.}.
	self assert: matches = (Dictionary newFrom: {y -> x. z -> x}).! !

!UnificationTest methodsFor: 'testing' stamp: 'fbs 1/4/2012 21:36'!
testVariablesInsideCollectionsUnify
	| matches |
	matches := {1. 2. #x asVariable. 4} =? #(1 2 3 4).
	self assert: matches equals: (Dictionary newFromPairs: {#x asVariable. 3.}).
	
	matches := #(1 2 3 4) =? {1. 2. #x asVariable. 4}.
	self assert: matches equals: (Dictionary newFromPairs: {#x asVariable. 3.}).! !

TestCase subclass: #VariableTrackingUnionFindTest
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Unification-Tests'!

!VariableTrackingUnionFindTest methodsFor: 'fixtures' stamp: 'fbs 1/4/2012 20:26'!
makePartition
	^ VariableTrackingUnionFind
		usingArrayType: PersistentCollection
		partitioning: Dictionary new! !

!VariableTrackingUnionFindTest methodsFor: 'as yet unclassified' stamp: 'fbs 1/4/2012 16:25'!
testAddingNonVariableToVariableClassMakesNonVariableTheRepresentative
	| v |
	v := self makePartition.
	v := v union: #x asVariable with: 1.
	self assert: 1 equals: (v find: 1).
	self assert: 1 equals: (v find: #x asVariable).! !

!VariableTrackingUnionFindTest methodsFor: 'as yet unclassified' stamp: 'fbs 1/4/2012 16:25'!
testAddingVariableToClassMakesNonVariableTheRepresentative
	| v |
	v := self makePartition.
	v := v union: 1 with: #x asVariable.
	self assert: 1 equals: (v find: 1).
	self assert: 1 equals: (v find: #x asVariable).! !

!VariableTrackingUnionFindTest methodsFor: 'as yet unclassified' stamp: 'fbs 1/4/2012 16:26'!
testAddingVariableToVariableClassMakesOriginalVariableTheRepresentative
	| v |
	v := self makePartition.
	v := v union: #x asVariable with: #y asVariable.
	self assert: #y asVariable equals: (v find: #y asVariable).
	self assert: #y asVariable equals: (v find: #x asVariable).! !

Object subclass: #MetaVariable
	instanceVariableNames: 'name'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Unification'!

!MetaVariable commentStamp: 'fbs 5/24/2011 11:04' prior: 0!
A MetaVariable is a reified variable, that is, an object representing a variable. It is a placeholder in a structure that says "we don't know what the value is of this part of the structure, but we can give it a name."!

!MetaVariable class methodsFor: 'instance creation' stamp: 'fbs 5/24/2011 07:31'!
named: aSymbol
	^ self new name: aSymbol.! !

!MetaVariable methodsFor: 'comparing' stamp: 'fbs 1/4/2012 20:33'!
= anObject
	^ anObject isMetaVariable and: [name = anObject name]! !

!MetaVariable methodsFor: 'accessing' stamp: 'fbs 5/24/2011 22:06'!
children
	^ #().! !

!MetaVariable methodsFor: 'testing' stamp: 'fbs 5/24/2011 22:05'!
hasChildren
	"Even though unapplied a variable looks like it has a parameter, that's just its name."
	^ false.! !

!MetaVariable methodsFor: 'comparing' stamp: 'fbs 1/4/2012 20:34'!
hash
	"17's a prime."
	"Implementation copied from AlgebraicDataType >> #hash."
	^ self unapply inject: 17 into: [:acc :each | 17 * each hash].! !

!MetaVariable methodsFor: 'testing' stamp: 'fbs 1/4/2012 20:16'!
isMetaVariable
	^ true.! !

!MetaVariable methodsFor: 'accessing' stamp: 'fbs 5/24/2011 07:30'!
name
	^ name.! !

!MetaVariable methodsFor: 'private' stamp: 'fbs 5/24/2011 07:32'!
name: aSymbol
	name := aSymbol.! !

!MetaVariable methodsFor: 'printing' stamp: 'fbs 1/4/2012 20:17'!
printOn: aStream
	aStream nextPut: $(.
	aStream nextPutAll: ((self unapply collect: [:token | token printString])
		reduce: [:acc :each | acc , ' ', each]).
	aStream nextPut: $).! !

!MetaVariable methodsFor: 'destructuring' stamp: 'fbs 5/24/2011 07:34'!
unapply
	^ {#Variable. name.}! !

!MetaVariable methodsFor: 'unification' stamp: 'fbs 1/3/2012 15:30'!
unificationClosureWith: anObject in: termRelation
	^ termRelation union: self with: anObject.! !

!Object methodsFor: '*Unification' stamp: 'fbs 10/7/2011 07:43'!
=? anObject
	^ self unifyWith: anObject.! !

!Object methodsFor: '*Unification' stamp: 'fbs 1/4/2012 11:32'!
findSolutionWithUnifier: anMGU in: solutionFinder
	^ anMGU.! !

!Object methodsFor: '*Unification' stamp: 'fbs 5/25/2011 22:33'!
isMetaVariable
	^ false.! !

!Object methodsFor: '*Unification' stamp: 'fbs 7/23/2011 11:25'!
matchOneOf: aBlockAssociationCollection
	"The elements of aBlockAssociationCollection are associations between blocks.
	 Answer the evaluated value of the first association in aBlockAssociationCollection
	 whose evaluated key unifies with the receiver.  If no match is found, report an error."

	^ self matchOneOf: aBlockAssociationCollection otherwise: [self caseError].! !

!Object methodsFor: '*Unification' stamp: 'fbs 8/8/2011 19:07'!
matchOneOf: someAssociations otherwise: aBlock
	"The elements of aBlockAssociationCollection are associations between blocks.
	 Answer the evaluated value of the first association in aBlockAssociationCollection
	 whose evaluated key unifies with the receiver.  If no match is found, answer the result
	 of evaluating aBlock."
	^ (UnificationMatcher matcherUsing: someAssociations otherwise: aBlock) match: self.! !

!Object methodsFor: '*Unification' stamp: 'fbs 1/3/2012 15:28'!
unificationClosureWith: anObject in: termRelation
	anObject isMetaVariable
		ifTrue: [^ termRelation union: self with: anObject].

	^ (self = anObject)
		ifTrue: [termRelation union: self with: anObject]
		ifFalse: ["s ~= t" UnificationFailure signal]! !

!Object methodsFor: '*Unification' stamp: 'fbs 1/9/2012 09:28'!
unificationClosureWithAlgebraic: anObject in: termRelation
	^ self unificationClosureWithSequence: anObject unapply in: termRelation.! !

!Object methodsFor: '*Unification' stamp: 'fbs 1/4/2012 21:34'!
unificationClosureWithSequence: aSequence in: termRelation
	UnificationFailure signal.! !

!Object methodsFor: '*Unification' stamp: 'fbs 5/26/2011 20:42'!
unifyWith: anObject
	^ Unifier unify: self with: anObject.! !

Object subclass: #OrUnifier
	instanceVariableNames: 'left right'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Unification'!

!OrUnifier commentStamp: 'fbs 8/8/2011 22:31' prior: 0!
An OrUnifier is an object that allows an object to match one of two different objects:

OrUnifier
	left: (Node left: #x asVariable)
	right: (Node right: #x asVariable)
	
returns something that unifies with single-child Nodes.!

!OrUnifier class methodsFor: 'instance creation' stamp: 'fbs 8/8/2011 22:33'!
left: leftObject right: rightObject
	^ self new left: leftObject right: rightObject.! !

!OrUnifier methodsFor: 'private' stamp: 'fbs 8/8/2011 22:33'!
left: leftObject right: rightObject
	left := leftObject.
	right := rightObject.! !

!OrUnifier methodsFor: '*Unification' stamp: 'fbs 1/4/2012 20:31'!
unificationClosureWithAlgebraic: anObject in: termRelation
	[^ anObject unificationClosureWithAlgebraic: left in: termRelation]
		on: UnificationFailure do:
			[^ anObject unificationClosureWithAlgebraic: right in: termRelation]! !

!OrUnifier methodsFor: '*Unification' stamp: 'fbs 8/9/2011 07:36'!
unifyWith: anObject
	[^ left unifyWith: anObject]
		on: UnificationFailure do: [^ right unifyWith: anObject].! !

Object subclass: #SolutionFinder
	instanceVariableNames: 'acyclic visited partition'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Unification'!

!SolutionFinder commentStamp: 'fbs 1/4/2012 10:29' prior: 0!
Given an equivalence relation (partition) on some set of nodes, each class potentially containing one or more MetaVariables, construct a MostGeneralUnifier.!

!SolutionFinder class methodsFor: 'instance creation' stamp: 'fbs 1/4/2012 11:35'!
findSolutionFor: aVariableAvoidingUnionFind starting: anObject
	^ self new findSolutionFor: aVariableAvoidingUnionFind starting: anObject.! !

!SolutionFinder methodsFor: 'private' stamp: 'fbs 1/4/2012 14:23'!
findSolution: anObject using: anMGU
	| newUnifier schemaTerm |
	schemaTerm := partition find: anObject.
	(acyclic includes: schemaTerm) ifTrue: [^ anMGU "s is not part of a cycle."].
	(visited includes: schemaTerm) ifTrue: [UnificationFailure signal. "Cycle detected."].
	
	newUnifier := anMGU copy.
	visited add: schemaTerm.
	newUnifier addAll: (schemaTerm findSolutionWithUnifier: newUnifier in: self).
	visited remove: schemaTerm.
	
	acyclic add: schemaTerm.
	(partition varsAt: schemaTerm) do:
		[:x | (x ~= schemaTerm) ifTrue: [newUnifier at: x put: schemaTerm]].
	^ newUnifier.! !

!SolutionFinder methodsFor: 'unification' stamp: 'fbs 1/4/2012 11:45'!
findSolutionFor: aVariableAvoidingUnionFind starting: anObject
	partition := aVariableAvoidingUnionFind.
	acyclic := OrderedCollection new.
	visited := OrderedCollection new.
	^ self findSolution: anObject using: MostGeneralUnifier new.! !

Object subclass: #UnificationMatcher
	instanceVariableNames: 'cases otherwise'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Unification'!

!UnificationMatcher commentStamp: 'fbs 7/28/2011 22:15' prior: 0!
I provide a unification-based case-like statement.

I use memoisation as an optimisation.!

!UnificationMatcher class methodsFor: 'instance creation' stamp: 'fbs 7/28/2011 22:29'!
matcherUsing: aBlockAssociationCollection
	^ self new matcherUsing: aBlockAssociationCollection.! !

!UnificationMatcher class methodsFor: 'instance creation' stamp: 'fbs 7/28/2011 22:14'!
matcherUsing: aBlockAssociationCollection otherwise: aBlock
	^ self new matcherUsing: aBlockAssociationCollection otherwise: aBlock.! !

!UnificationMatcher methodsFor: 'applying' stamp: 'fbs 8/8/2011 22:11'!
match: anObject
	"Iterate through the cases. As we evaluate the key thunks, replace them with the evaluated values. Much uglier than the naive implementation, but MUCH faster."
	| newCases result unmatch |
	result := unmatch := Object new.
	
	"As we try unify against the key thunk, build up a new Dictionary with the unifying key thunk memoised."
	newCases := OrderedCollection new.
	cases associationsDo:
		[:assoc | | mgu newKey |
			[newKey := assoc key isBlock ifTrue: [assoc key value] ifFalse: [assoc key].
			mgu := newKey =? anObject.
			result = unmatch ifTrue: [result := assoc value cull: mgu].
			newCases
				add: newKey;
				add: assoc value.] on: UnificationFailure do: ["nothing"]].

	cases := Dictionary newFromPairs: newCases.
	
	^ result == unmatch
		ifTrue: [otherwise value]
		ifFalse: [result].! !

!UnificationMatcher methodsFor: 'initialize-release' stamp: 'fbs 1/8/2012 22:05'!
matcherUsing: aBlockAssociationCollection
	self matcherUsing: aBlockAssociationCollection otherwise: [self caseError].! !

!UnificationMatcher methodsFor: 'initialize-release' stamp: 'fbs 8/8/2011 22:22'!
matcherUsing: aBlockAssociationCollection otherwise: aBlock
	cases := aBlockAssociationCollection asOrderedCollection.
	otherwise := aBlock.! !

Object subclass: #Unifier
	instanceVariableNames: 'termRelation'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Unification'!

!Unifier commentStamp: 'fbs 7/23/2011 11:32' prior: 0!
I attempt to unify two structures, that is, find mappings between the MetaVariables (if any) within the two structures such that the structures represent the same structure.

I return a Dictionary containing the most general unifier (which might be empty if the two structures are identical), or I raise a UnificationFailure.!

!Unifier class methodsFor: 'instance creation' stamp: 'fbs 5/25/2011 07:12'!
unify: left with: right
	^ self new unify: left with: right.! !

!Unifier methodsFor: 'evaluating' stamp: 'fbs 1/4/2012 20:26'!
unify: s with: t
	termRelation := VariableTrackingUnionFind usingArrayType: PersistentCollection partitioning: Dictionary new.
	termRelation := (termRelation find: s) unificationClosureWith: (termRelation find: t) in: termRelation.
	^ SolutionFinder findSolutionFor: termRelation starting: s.! !

PersistentUnionFind subclass: #VariableTrackingUnionFind
	instanceVariableNames: 'vars'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Unification'!

!VariableTrackingUnionFind commentStamp: 'fbs 1/4/2012 20:42' prior: 0!
A VariableAvoidingUnionFind is a persistent union-find structure that also tracks MetaVariables.

In particular, it maintains the invariants of a _schema function_ mapping equivalence classes to terms: a schema function S is a function such that
* S(C) elem C,
* S(C) is a MetaVariable only if every element in C is a MetaVariable.!

!VariableTrackingUnionFind methodsFor: 'accessing' stamp: 'fbs 1/4/2012 13:31'!
at: anObject addVar: aMetaVariable
	| representative |
	representative := self find: anObject.
	^ vars at: representative put: ((vars at: representative) , {aMetaVariable})! !

!VariableTrackingUnionFind methodsFor: 'private' stamp: 'fbs 1/3/2012 16:02'!
copyWithRanks: anArray parents: anotherArray
	^ self class new ranks: anArray parents: anotherArray vars: vars.! !

!VariableTrackingUnionFind methodsFor: 'initialize-release' stamp: 'fbs 1/3/2012 16:10'!
initialize
	super initialize.
	vars := PersistentCollection initially: (DictionaryWithDefault wrapping: Dictionary new default: [:unused | Set new]).! !

!VariableTrackingUnionFind methodsFor: 'private' stamp: 'fbs 1/3/2012 16:03'!
ranks: anArray parents: anotherArray vars: aThirdArray
	super ranks: anArray parents: anotherArray.
	vars := aThirdArray.! !

!VariableTrackingUnionFind methodsFor: 'private' stamp: 'fbs 1/4/2012 16:28'!
storeSchemaTermFor: anObject as: anotherObject
	"The schema function is a function from terms to equivalence classes such that:
	* sigma(C) elem C,
	* sigma(C) is a variable only if every c in C is a variable."
	(anObject isMetaVariable and: [anotherObject isMetaVariable])
		ifTrue:
			[vars := self at: anotherObject addVar: anObject.
			^ super storeSchemaTermFor: anObject as: anotherObject].

	anotherObject isMetaVariable
		ifTrue:
			[vars := self at: anObject addVar: anotherObject.
			^ super storeSchemaTermFor: anotherObject as: anObject].

	anObject isMetaVariable
		ifTrue:
			[vars := self at: anotherObject addVar: anObject].
	^ super storeSchemaTermFor: anObject as: anotherObject.! !

!VariableTrackingUnionFind methodsFor: 'accessing' stamp: 'fbs 1/3/2012 16:01'!
varsAt: anObject
	^ vars at: anObject! !

!AlgebraicDataType methodsFor: '*Unification' stamp: 'fbs 1/4/2012 11:31'!
findSolutionWithUnifier: anMGU in: solutionFinder
	| newUnifier |
	newUnifier := anMGU copy.
	self children do:
		[:each | | d |
		d := solutionFinder findSolution: each using: anMGU.
		newUnifier addAll: d].
	^ newUnifier.! !

!AlgebraicDataType methodsFor: '*Unification' stamp: 'fbs 5/25/2011 22:31'!
isMetaVariable
	^ self unapply first = #Variable.! !

!AlgebraicDataType methodsFor: '*Unification' stamp: 'fbs 8/9/2011 07:32'!
or: anObject
	^ OrUnifier left: self right: anObject! !

!AlgebraicDataType methodsFor: '*Unification' stamp: 'fbs 1/4/2012 20:20'!
unificationClosureWith: anObject in: termRelation
	anObject isMetaVariable
		ifTrue: [^ termRelation union: self with: anObject].

	^ anObject unificationClosureWithAlgebraic: self in: termRelation.! !

!AlgebraicDataType methodsFor: '*Unification' stamp: 'fbs 12/30/2011 22:14'!
unificationClosureWithAlgebraic: anObject in: termRelation
	"s = f(s1, ..., sm), t = g(t1, ..., tn) where m, n > 0"
	| partition |
	(self type = anObject type) ifFalse: [UnificationFailure signal. "f ~= g"].
	(self arity = anObject arity) ifFalse: [UnificationFailure signal. "m ~= n"].
	partition := termRelation.
	self children with: anObject children do:
		[:si :ti | partition := (partition find: si) unificationClosureWith: (partition find: ti) in: partition].
	^ partition.! !

AlgebraicDataType subclass: #VariadicTerm
	instanceVariableNames: 'subterms'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Unification-Tests'!

!VariadicTerm commentStamp: 'fbs 5/29/2011 16:47' prior: 0!
A general term that may contain any (variable) number of subterms. It's mainly just for testing.!

!VariadicTerm class methodsFor: 'instance variables' stamp: 'fbs 5/27/2011 20:26'!
with: aCollection
	^ self new with: aCollection.! !

!VariadicTerm methodsFor: 'destructuring' stamp: 'fbs 5/27/2011 20:27'!
unapply
	^ #Term asOrderedCollection, subterms.! !

!VariadicTerm methodsFor: 'private' stamp: 'fbs 5/27/2011 20:26'!
with: aCollection
	subterms := aCollection! !
